@@ -3,7 +3,7 @@ package App::ClusterSSH;
 use 5.008.004;
 use warnings;
 use strict;
-use version; our $VERSION = version->new('4.04');
+use version; our $VERSION = version->new('4.05');
 
 use Carp qw/cluck :DEFAULT/;
 
@@ -37,6 +37,7 @@ use Net::hostent;
 use Sys::Hostname;
 use English;
 use Socket;
+use File::Path qw(make_path);
 
 # Notes on general order of processing
 #
@@ -510,6 +511,39 @@ SWITCH: {
     return $self;
 }
 
+sub substitute_macros {
+    my ( $self, $svr, $text ) = @_;
+
+    return $text unless ( $self->config->{macros_enabled} eq 'yes' );
+
+    {
+        my $macro_servername = $self->config->{macro_servername};
+        ( my $servername = $svr ) =~ s/\s+//;
+        $text =~ s!$macro_servername!$servername!xsmg;
+    }
+    {
+        my $macro_hostname = $self->config->{macro_hostname};
+        my $hostname       = $servers{$svr}{givenname};
+        $text =~ s!$macro_hostname!$hostname!xsmg;
+    }
+    {
+        my $macro_username = $self->config->{macro_username};
+        my $username       = $servers{$svr}{username};
+        $username ||= getpwuid($UID);
+        $text =~ s!$macro_username!$username!xsmg;
+    }
+    {
+        my $macro_newline = $self->config->{macro_newline};
+        $text =~ s!$macro_newline!\n!xsmg;
+    }
+    {
+        my $macro_version = $self->config->{macro_version};
+        $text =~ s/$macro_version/$VERSION/xsmg;
+    }
+
+    return $text;
+}
+
 sub send_text($@) {
     my $self = shift;
     my $svr  = shift;
@@ -518,34 +552,7 @@ sub send_text($@) {
     $self->debug( 2, "servers{$svr}{wid}=$servers{$svr}{wid}" );
     $self->debug( 3, "Sending to '$svr' text:$text:" );
 
-    # command macro substitution
-    if ( $self->config->{macros_enabled} eq 'yes' ) {
-
-        # $svr contains a trailing space here, so ensure its stripped off
-        {
-            my $macro_servername = $self->config->{macro_servername};
-            my $servername       = $svr;
-            $servername =~ s/\s+//;
-            $text =~ s/$macro_servername/$servername/xsmg;
-        }
-        $text =~ s/%h/hostname()/xsmeg;
-
-        # use connection username, else default to current username
-        {
-            my $macro_username = $self->config->{macro_username};
-            my $username       = $servers{$svr}{username};
-            $username ||= getpwuid($UID);
-            $text =~ s/$macro_username/$username/xsmg;
-        }
-        {
-            my $macro_newline = $self->config->{macro_newline};
-            $text =~ s/$macro_newline/\n/xsmg;
-        }
-        {
-            my $macro_version = $self->config->{macro_version};
-            $text =~ s/$macro_version/$VERSION/xsmg;
-        }
-    }
+    $text = $self->substitute_macros( $svr, $text );
 
     foreach my $char ( split( //, $text ) ) {
         next if ( !defined($char) );
@@ -578,7 +585,6 @@ sub send_text($@) {
                     'name'        => $event,
                     'detail'      => $code,
                     'state'       => $state,
-                    'time'        => time(),
                     'event'       => $servers{$svr}{wid},
                     'root'        => $xdisplay->root(),
                     'same_screen' => 1,
@@ -724,6 +730,12 @@ sub open_client_windows(@) {
           # Since this is the child, we can mark any server unresolved without
           # affecting the main program
             $servers{$server}{realname} .= "==" if ( !$realname );
+
+            # copy and amend the config provided to the helper script
+            my $local_config = $self->config;
+            $local_config->{command} = $self->substitute_macros( $server,
+                $local_config->{command} );
+
             my $exec = join( ' ',
                 $self->config->{terminal},
                 $color,
@@ -1681,11 +1693,14 @@ sub key_event {
                 $self->debug( 3, "matched combo" );
                 if ( $event eq "KeyRelease" ) {
                     $self->debug( 2, "Received hotkey: $hotkey" );
-                    $self->send_text_to_all_servers('%s')
+                    $self->send_text_to_all_servers(
+                        $self->config->{macro_servername} )
                         if ( $hotkey eq "key_clientname" );
-                    $self->send_text_to_all_servers('%h')
+                    $self->send_text_to_all_servers(
+                        $self->config->{macro_hostname} )
                         if ( $hotkey eq "key_localname" );
-                    $self->send_text_to_all_servers('%u')
+                    $self->send_text_to_all_servers(
+                        $self->config->{macro_username} )
                         if ( $hotkey eq "key_username" );
                     $self->add_host_by_name()
                         if ( $hotkey eq "key_addhost" );
@@ -1723,7 +1738,6 @@ sub key_event {
                     'name'        => $event,
                     'detail'      => $keycode,
                     'state'       => $state,
-                    'time'        => time(),
                     'event'       => $servers{$_}{wid},
                     'root'        => $xdisplay->root(),
                     'same_screen' => 1,
@@ -2031,14 +2045,32 @@ sub run {
     $self->cluster->get_tag_entries( split /,/,
         $self->config->{extra_tag_file} || '' );
 
-    if ( $self->options->list ) {
-        print( 'Available cluster tags:', $/ );
-        print "\t", $_, $/ foreach ( sort( $self->cluster->list_tags ) );
+    if ( defined $self->options->list ) {
+        my $eol = $self->options->quiet ? ' ' : $/;
+        my $tab = $self->options->quiet ? ''  : "\t";
+        if ( !$self->options->list ) {
+            print( 'Available cluster tags:', $/ )
+                unless ( $self->options->quiet );
+            print $tab, $_, $eol
+                foreach ( sort( $self->cluster->list_tags ) );
 
-        my @external_clusters = $self->cluster->list_external_clusters;
-        if (@external_clusters) {
-            print( 'Available external command tags:', $/ );
-            print "\t", $_, $/ foreach ( sort(@external_clusters) );
+            my @external_clusters = $self->cluster->list_external_clusters;
+            if (@external_clusters) {
+                print( 'Available external command tags:', $/ )
+                    unless ( $self->options->quiet );
+                print $tab, $_, $eol foreach ( sort(@external_clusters) );
+                print $/;
+            }
+        }
+        else {
+            print 'Tag resolved to hosts: ', $/
+                unless ( $self->options->quiet );
+            @servers = $self->resolve_names( $self->options->list );
+
+            foreach my $svr (@servers) {
+                print $tab, $svr, $eol;
+            }
+            print $/;
         }
 
         $self->debug(
@@ -2071,7 +2103,8 @@ sub run {
     $self->debug( 2, "Capture map events" );
     $self->capture_map_events();
 
-    $self->debug( 0, 'Opening to: ', join( ' ', @servers ) ) if (@servers);
+    $self->debug( 0, 'Opening to: ', join( ' ', @servers ) )
+        if ( @servers && !$self->options->quiet );
     $self->open_client_windows(@servers);
 
     # Check here if we are tiling windows.  Here instead of in func so
@@ -2226,6 +2259,8 @@ the code until this time.
 
 =item show_history
 
+=item substitute_macros
+
 =item terminate_host
 
 =item toggle_active_state
